"""Module for generating and playing sounds.

This module contains the following files:

    sounds.py : Defines classes for generating sounds
    jackclient.py : Define the interface to the jack client
    pyoserver.py : Defines the interface to the pyo server

The use of pyoserver is discouraged in favor of jackclient. This is
controlled by the pref `AUDIOSERVER`.
"""
