"""
Tests that will hopefully become unnecessary as testing coverage increases -- just testing that we can import things that we expect to be able to!
"""

def test_import_stim():
    from autopilot.stim import Stim